;;; patcher-diff.el --- Diff utilities

;; Copyright (C) 2008, 2009, 2010 Didier Verna
;; Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007 Didier Verna

;; Author:        Didier Verna <didier@xemacs.org>
;; Maintainer:    Didier Verna <didier@xemacs.org>
;; Created:       Sun Feb 14 16:58:16 2010
;; Last Revision: Fri Dec  2 22:11:00 2011
;; Keywords:      maint


;; This file is part of Patcher.

;; Patcher is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License version 3,
;; as published by the Free Software Foundation.

;; Patcher is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;;; Commentary:

;; Contents management by FCM version 0.1.


;;; Code:

(require 'cl)

(eval-when-compile (require 'patcher-cutil))
(require 'patcher-util)
(require 'patcher-project)
(require 'patcher-instance)
(require 'patcher-source)
(require 'patcher-change-log)



;; ==========================================================================
;; Utilities
;; ==========================================================================

(defun* patcher-run-after-diff-hook
    (project min max
     &aux (after-diff-hook (patcher-project-option project :after-diff-hook)))
  ;; Run PROJECT's after-diff-hook in current buffer, between MIN and MAX.
  (when after-diff-hook
    (patcher-with-progression "Running the after-diff hook"
      (mapcar (lambda (func)
		(funcall func min max))
	      after-diff-hook))))




;; ==========================================================================
;; Diff creation and parsing
;; ==========================================================================

(defun patcher-parse-diff (project min max)
  ;; Parse PROJECT's diff output in current buffer between MIN and MAX.
  ;; For each diffed file, create extents with the following properties:
  ;; 'patcher-project = PROJECT, and
  ;; 'patcher-change-log = <absolute filename> for ChangeLog files, or
  ;; 'patcher-source = <absolute filename> for source files.
  (patcher-with-progression "Parsing the diff output"
    (save-excursion
      (goto-char min)
      (let* ((diff-header (patcher-project-option project :diff-header t))
	     (change-logs-updating (patcher-project-option project
				     :change-logs-updating))
	     (regexp         (nth 0 diff-header))
	     (old-file-match (nth 1 diff-header))
	     (new-file-match (nth 2 diff-header))
	     change-log
	     old-file old-absfile
	     new-file new-absfile
	     beg end)
	(while (re-search-forward regexp max t)
	  (setq beg (match-beginning 0))
	  (setq old-file (match-string old-file-match))
	  (setq new-file (match-string new-file-match))
	  (if (string= old-file "/dev/null")
	      (setq old-file nil)
	    (setq old-absfile (expand-file-name old-file default-directory)))
	  (if (string= new-file "/dev/null")
	      (setq new-file nil)
	    (setq new-absfile (expand-file-name new-file default-directory)))
	  (setq end (or (save-excursion
			  (when (re-search-forward regexp max t)
			    (match-beginning 0)))
			max))
	  ;; #### NOTE: the extent properties below are set in relation with
	  ;; NEW-FILE, unless we're facing a deletion, in which case we use
	  ;; OLD-FILE instead.
	  (when change-logs-updating
	    (setq change-log (patcher-locate-change-log project
			       (or new-absfile old-absfile))))
	  (set-extent-properties (make-extent beg end)
				 `(start-open t
				   duplicable t
				   patcher-project ,project
				   ,(if (and change-log
					     (string= change-log
						      (or new-absfile
							  old-absfile)))
					'patcher-change-log
					'patcher-source)
				   ,(or new-absfile old-absfile))))))))

(patcher-define-error 'diff
  "Patcher diff error."
  'process)

(put 'patcher-diff 'lisp-indent-function 2)
(defun* patcher-diff
    (project command &key (erase t) (progression "Diff'ing the project"))
  ;; Create a PROJECT diff with COMMAND in process buffer.
  ;; If ERASE (the default), erase the process buffer first.
  ;; Otherwise, insert the diff at current point.
  ;; Use PROGRESSION message to make people wait.
  ;; After the diff is created, run the diff line filter, the after diff hook
  ;; and parse the diff output.
  ;; Throw a PATCHER-DIFF error if COMMAND fails.
  (multiple-value-bind (min max)
      (patcher-condition-case nil
	  (patcher-call-command project command
	    :erase erase
	    :progression progression
	    :ignore-exit-status
	    (patcher-project-option project :ignore-diff-status))
	(process
	 (patcher-error 'diff command)))
    (with-current-buffer (patcher-project-process-buffer project)
      ;; We need a marker in case the hook modifies the diff output.
      (goto-char max)
      (setq max (point-marker))
      (let ((diff-line-filter (patcher-project-option project
				:diff-line-filter)))
	(when diff-line-filter
	  (operate-on-matching-lines diff-line-filter t nil min max)))
      (patcher-run-after-diff-hook project min max)
      (patcher-parse-diff project min max))))

(defun patcher-diff-change-logs (project)
  ;; Create PROJECT's ChangeLog diff.
  ;; This function assumes that ChangeLog files are already known and filled
  ;; in properly.
  (patcher-save-change-logs project)
  (patcher-diff project (patcher-command project
			  (or (patcher-project-option project
				:change-logs-diff-command)
			      (patcher-project-diff-command project))
			  (mapcar #'patcher-file-relative-name
				  (patcher-project-change-logs project)))
    :progression "Generating the ChangeLogs diff")
  (patcher-detect-missing-change-logs project))

(defun* patcher-convert-change-logs-diff
    (project
     ;; Here so that it triggers an error if nil, before doing anything else.
     &aux (change-logs-diff-command (patcher-project-option project
				      :change-logs-diff-command t))
	  (buffer (patcher-project-process-buffer project))
	  change-log)
  ;; Convert PROJECT's ChangeLog diffs in process buffer.
  ;; Conversion is done directly in place with the ChangeLog-specific diff
  ;; command.
  (patcher-with-progression "Converting the ChangeLogs diff"
    (with-current-buffer buffer
      ;; #### WARNING: deleting extents from within mapcar-extents seems
      ;; unsafe. Besides, if I modify the extents'contents instead of deleting
      ;; and recreating them, map(car)-extents goes into an infinite loop, on
      ;; all extents over and over again.
      (dolist (extent (patcher-change-log-extents))
	(setq change-log (extent-property extent 'patcher-change-log))
	(goto-char (extent-start-position extent))
	(patcher-delete-extent-and-region extent)
	(patcher-diff project
	    (patcher-command project change-logs-diff-command
			     (list (patcher-file-relative-name change-log)))
	  :erase nil
	  :progression (format "Processing %s" change-log))))))

(defun patcher-diff-all (project)
  ;; Create PROJECT's global diff with both ChangeLog and source files.
  ;; This function assumes that ChangeLog files are already known and filled
  ;; in properly.
  (patcher-save-sources project)
  (patcher-save-change-logs project)
  (patcher-diff project
      (patcher-command project (patcher-project-diff-command project) t)
    :progression "Generating global diff")
  (patcher-detect-inconsistent-sources project)
  (patcher-detect-inconsistent-change-logs project)
  (when (patcher-project-option project :change-logs-diff-command)
    (patcher-convert-change-logs-diff project)))

(defun* patcher-diff-specification
    (project
     &aux (buffer (patcher-project-process-buffer project))
	  sources
	  change-logs)
  ;; Diff PROJECT's specification and deduce source and ChangeLog files.
  ;; Return 2 values: the lists of involved source and ChangeLog files.
  (patcher-save-sources project)
  (patcher-save-change-logs project)
  (patcher-diff project (patcher-command project
			  (patcher-project-diff-command project)
			  (patcher-project-specification project)))
  (setq sources (patcher-sources buffer))
  (unless sources
    (patcher-error "Your source files do not differ from the archive."))
  (when (patcher-project-option project :change-logs-updating)
    (patcher-with-progression "Detecting ChangeLog files"
      ;; Note that while we can rely on the diff to deduce the source files,
      ;; this is not possible for ChangeLog files, even when the updating is
      ;; manual, because they might not be there (for instance if
      ;; `specification' is explicit about files). So all in all, the only
      ;; safe way to deduce ChangeLog files is to compute them from the source
      ;; diffs.
      (dolist (source sources)
	(let ((change-log (patcher-locate-change-log project source)))
	  (unless (member change-log change-logs)
	    (patcher-endpush change-log change-logs))))))
  (values sources change-logs))




;; ==========================================================================
;; Diff insertion
;; ==========================================================================

(globally-declare-boundp 'font-lock-always-fontify-immediately)

(put 'patcher-insert-diff 'lisp-indent-function 1)
(defun* patcher-insert-diff
    (project point &optional (kind :sources)
		   &aux (buffer (patcher-project-process-buffer project))
			(function (patcher-project-option project
				    :diff-prologue-function)))
  ;; Insert PROJECT's diff at POINT in current buffer.
  ;; The diff comes from the process buffer and can be of different KINDs:
  ;; - :sources (the default) means to filter out only source diffs,
  ;; - :change-logs means to filter out only ChangeLog diffs,
  ;; - :mixed means to keep both (the whole process buffer is used here).
  ;; Prepend the diff with an appropriate prologue.
  ;; Depending on KIND, create a patcher-diff or patcher-change-logs extent
  ;; around insertion.
  (save-excursion
    (goto-char point)
    (when function
      (funcall function
	       (patcher-project-command-name project)
	       kind
	       :source-files (patcher-project-sources-string project)
	       :change-log-files (patcher-project-change-logs-string project)
	       :source-diff (patcher-command project
			      (patcher-project-diff-command project))
	       :change-log-diff (patcher-command project
				  (patcher-project-option project
				    :change-logs-diff-command))))
    (let ((font-lock-always-fontify-immediately t))
      (insert (ecase kind
		(:sources
		 (patcher-source-contents buffer))
		(:change-logs
		 (patcher-change-log-contents buffer))
		(:mixed
		 (buffer-substring nil nil buffer)))))
    (set-extent-properties (make-extent point (point))
			   `(start-open t
			     duplicable t
			     ,(if (eq kind :change-logs)
				  'patcher-change-logs
				  'patcher-diff) t))))


(provide 'patcher-diff)

;;; patcher-diff.el ends here
